#!/usr/bin/env python

import subprocess
import re
from collections import namedtuple



class VMStat(object):
	#
	# Some general facts about wired memory and kernel memory allocation:
	# - memory can be wired as a result of high level kernel memory allocation
	#   or as a result of wiring userland pages via mlock(2) and similar
	#   or as a result of direct use of low level memory page management routines
	# - the latter are used for example to create page tables
	# - there is no accounting for low level wiring
	# - there is no centralized accounting for userland wirings
	# - high level kernel memory allocation can be done in two ways: via malloc(9) and via uma(9)
	# - there is accounting of malloc memory allocations per malloc tag/type
	# - there is accounting of uma allocation per uma zone
	# - malloc memory allocations smaller or equal to page size are actually backed by special uma zones
	# - so such allocations are accounted against both the corresponding malloc tag and a zone correspondning to allocation size
	# - uma zones may cache some memory for re-use, so both actively used and cached items are accounted per zone
	#
	# Some facts on ZFS and ZFS ARC memory composition:
	# - buffers for storing copies of on-disk data can be allocated either via malloc or via uma
	# - ARC size accounts for the following sub-categories:
	#   * data buffers accounted as "data"
	#     - data buffers hold both data and metadata
	#   * objects providing structured representation of data (such as dnodes) acccounted as "other"
	#     - some of these allocations are via uma and some are via malloc
	#   * objects used for ARC management accounted as "headers"
	#     - these allocations are done via uma
	#   * objects used for L2ARC management acounted as "l2 headers"
	#     - these allocations are done via malloc
	# - ZFS makes a lot of other short term and long term allocations
	# - these allocations are done both via uma and malloc
	# - some examples: znodes, zap entries, range lock internals
	# - part of data buffers belong to ARC "data", but others can be used for temporary storage and are not accounted there
	# - also data buffers that hold dnode 'bonus' data are accounted as ARC "other"
	#
	# Calculations below try to untange the overlapping accounting.
	#

	def __init__(self):
		self.vmstat_m_re = re.compile(r'^(.*) ([0-9]+) +([0-9]+)K +[^ ]+ +[0-9]+ +([0-9,]+)?$')
		self.vmstat_z_re = re.compile(r'^([^:]*):\s*([0-9]+),\s*[0-9]+,\s*([0-9]+),\s*([0-9]+),\s*[0-9]+,\s*[0-9]+,\s*[0-9]+$')
		self.zfs_zio_buf_re = re.compile(r'zio_data_buf_\d+|zio_buf_\d|zfs/zio_buf')
		self.zfs_arc_other_re = re.compile('dnode_t|dmu_buf_impl_t')
		self.zfs_arc_header_re = re.compile('arc_buf_t|arc_buf_hdr_t')

		self.malloc_zones = ( '16', '32', '64', '128', '256', '512', '1024', '2048', '4096', '8192', '16384', '32768', '65536' )

		self.StatRecord = namedtuple('StatRecord', 'name, n_items, n_bytes, is_zone, is_free')
		self.stats = []

		self.page_size = self.sysctl('vm.stats.vm.v_page_size')
		self.use_uma = self.sysctl('vfs.zfs.zio.use_uma')


	def sysctl(self, name):
		output = subprocess.check_output('sysctl -n %s' % (name), shell=True)
		return int(output.strip())


	def vmstat_m(self):
		cmd = subprocess.Popen('vmstat -m', shell=True, stdout=subprocess.PIPE)
		for line in cmd.stdout:
			match = self.vmstat_m_re.match(line)
			if match:
				name = match.group(1).strip()
				n_items = int(match.group(2))
				n_bytes = int(match.group(3)) * 1024
				#sizes = match.group(4)
				if n_items != 0:
					record = self.StatRecord(name, n_items, n_bytes, False, False)
					self.stats.append(record)

			
	def vmstat_z(self):
		cmd = subprocess.Popen('vmstat -z', shell=True, stdout=subprocess.PIPE)
		for line in cmd.stdout:
			match = self.vmstat_z_re.match(line)
			if match:
				name = match.group(1).strip()
				size = int(match.group(2))
				used = int(match.group(3))
				free = int(match.group(4))
				if not name in self.malloc_zones and used != 0:
					record = self.StatRecord(name, used, used * size, True, False)
					self.stats.append(record)
				if free != 0:
					record = self.StatRecord(name, free, free * size, True, True)
					self.stats.append(record)


	def getStats(self):
		wired_pages = self.sysctl('vm.stats.vm.v_wire_count')
		wired = wired_pages * self.page_size
		arc_size = self.sysctl('kstat.zfs.misc.arcstats.size')
		arc_data_size = self.sysctl('kstat.zfs.misc.arcstats.data_size')
		arc_header_size = self.sysctl('kstat.zfs.misc.arcstats.hdr_size')
		arc_l2header_size = self.sysctl('kstat.zfs.misc.arcstats.l2_hdr_size')
		arc_other_size = self.sysctl('kstat.zfs.misc.arcstats.other_size')
		self.vmstat_m()
		self.vmstat_z()
		self.stats.sort(key = lambda item : item.n_bytes)
		used = 0
		uma = 0
		cache = 0
		zfs = 0
		zio = 0
		arc_other = 0
		other = 0
		solaris = 0
		for item in self.stats:
			if item.is_free:
				print("%u bytes, %s [cache]" % (item.n_bytes, item.name))
				cache += item.n_bytes
				continue

			print("%u bytes, %s" % (item.n_bytes, item.name))
			used += item.n_bytes
			if item.is_zone:
				uma += item.n_bytes

			if item.name.startswith("zio_data_buf_") or item.name.startswith("zio_buf_"):
				zio += item.n_bytes
			elif item.name.startswith("zfs_") or item.name.startswith('zio_') or item.name in ('range_seg_cache', 'sa_cache', 'lz4_ctx', 'zil_lwb_cache') :
				zfs += item.n_bytes
			elif item.name in ( 'dnode_t', 'dmu_buf_impl_t', 'arc_buf_t', 'arc_buf_hdr_t' ):
				arc_other += item.n_bytes
			elif item.name == 'solaris':
				solaris += item.n_bytes
			else:
				other += item.n_bytes


		print("")
		print("Total memory in uma cache: %u" % (cache))
		print("Total memory used in uma: %u" % (uma))
		print("Total memory used in malloc: %u" % (used - uma))
		print("Total kernel memory allocated via malloc and uma: %u" % (used + cache))
		print("Total wired memory: %u" % (wired))
		print("ARC size: %u" % (arc_size))

		print("")
		print("ARC data: %u" % (arc_data_size))
		print("solaris: %u" % (solaris))
		print("zio: %u" % (zio))
		print("zfs: %u" % (zfs))
		print("arc_other (derived): %u" % (arc_other + zio - arc_data_size))
		print("arc_other (header + other stats): %u" % (arc_header_size + arc_l2header_size + arc_other_size))

		print("")
		print("%u%% (%u bytes) of wired memory is not allocated via malloc and uma" % (100 * (wired - used - cache) / wired, wired - used - cache))
		print("%u%% of wired memory is ARC" % (100 * arc_size / wired))


		if self.use_uma != 0:
			print("%u%% of wired memory are non-ARC ZFS, DTrace and Solaris compatibility allocations" % (100 * (solaris + zfs + zio - arc_data_size) / wired))
		else:
			print("%u%% of wired memory are non-ARC ZFS, DTrace and Solaris compatibility allocations" % (100 * (solaris - arc_data_size) / wired))

		print("%u%% of wired memory are other kernel allocations" % (100 * other / wired))
		print("%u%% of wired memory is used by uma caches" % (100 * cache / wired))

		print("")
		print("%u%% of uma memory is used by uma caches" % (100 * cache / (uma + cache)))


if __name__ == "__main__":
	vmstat = VMStat()
	vmstat.getStats()
